import { CollectionView } from "@mescius/wijmo";
import * as wjcGrid from "@mescius/wijmo.grid";
import { FlexGrid as wjFlexGrid } from "@mescius/wijmo.grid";
import { FlexGridXlsxConverter } from "@mescius/wijmo.grid.xlsx";
import React, { useCallback, useEffect } from "react";
import { getData } from "../../data/getData";
import { finishGenerateGridExportFile } from "../../store/grid/GridSlice";
import { useAppDispatch, useAppSelector } from "../../store/store";
import { WorkbookStyle } from "@mescius/wijmo.xlsx";
import { renderToString } from 'react-dom/server';

interface IWijmoGridExporterProps {
  /** The id of the grid to handle exports for. */
  gridId: string,
}

/**
 * Handles the redux action for exporting grids, checks if the grid is equal to the one passed in as
 * prop, and then exports that grid to XLSX.
 */
const WijmoGridExporter = React.forwardRef<wjFlexGrid, IWijmoGridExporterProps>((props, forwardedRef) => {
  const gridId = props.gridId;
  const gridIdToExport = useAppSelector(store => store.grid.generateExportFileOp?.data);
  const dispatch = useAppDispatch();

  const exportGridToExcel = useCallback(async () => {
    const grid = (forwardedRef as any)?.current;
    if (grid) {
      const values = await getData();

      // Quick hack to convert all true/false to yes/no for Excel export.
      if (Array.isArray(values)) {
        values.forEach((item: any) => {
          Object.entries(item).forEach(entry => {
            if (entry[1] === true) {
              item[entry[0]] = "Yes";
            } else if (entry[1] === false) {
              item[entry[0]] = "No";
            }
          });
        });
      }

      try {
        // Make a cloned grid.
        const clonedGrid = getClonedGrid(grid, values, false);

        // Export to excel.
        return FlexGridXlsxConverter.saveAsync(clonedGrid,
          {
            includeColumnHeaders: true,
            includeStyles: true,
            formatItem: (args) => {
              if (args.panel.cellType === wjcGrid.CellType.ColumnHeader) {
                args.xlsxCell.style = args.xlsxCell.style || {};
                args.xlsxCell.style.font = args.xlsxCell.style.font || {};
                args.xlsxCell.style.font.bold = true;
              }

              if (args.panel.cellType === wjcGrid.CellType.Cell) {
                let col = args.panel.columns[args.col];
                if (col.header == 'Issues') { // used header prop as binding was set to null, you can use binding or assign 'name' prop to the column
                  let xlsxCell = args.xlsxCell;
                  xlsxCell.value = args.cell.textContent;
                  if (!xlsxCell.style) {
                    xlsxCell.style = new WorkbookStyle();
                  }
                  xlsxCell.style.fill = { color: 'pink' }; // applied the style on complete cell, as background for text in a cell is not supported in Excel
                }
              }
            }
          },
          `${gridId}_Grid_Exported.xlsx`);
      } catch { }
    }
  }, [gridId, forwardedRef]);

  useEffect(() => {
    const asyncExport = async () => {
      if (gridIdToExport === gridId) {
        try {
          await exportGridToExcel();
        } catch (err) {
          alert(err);
        } finally {
          dispatch(finishGenerateGridExportFile());
        }
      }
    };

    asyncExport();
  }, [gridIdToExport, gridId, exportGridToExcel, dispatch]);

  // Don't render anything to the screen.
  return null;
});

export default WijmoGridExporter;

function getClonedGrid(grid: any, sourceCollection: any[], cloneFilter: boolean) {
  const collViewSettings = {
    sortDescriptions: grid.collectionView.sortDescriptions,
    groupDescriptions: grid.collectionView.groupDescriptions,
    filter: cloneFilter ? grid.collectionView.filter : undefined,
  };

  if (!cloneFilter) {
    delete collViewSettings.filter;
  }

  var sourceData = new CollectionView(sourceCollection, collViewSettings);

  var clone = new wjcGrid.FlexGrid(document.createElement('div'), {
    itemsSource: sourceData,
    columnLayout: grid.columnLayout
  });

  grid.columns.forEach((col: any) => {
    if (col.dataMap) {
      clone.columns[col.index].dataMap = col.dataMap;
    }
  });

  clone.formatItem.addHandler((s, e) => {
    if (s.cells !== e.panel) {
      return;
    }

    let templ = grid.cells.columns[e.col]['$__cellTemplCell'];
    if (templ) {
      // compile templates
      let context = {
        row: e.row,
        col: e.col,
        item: e.panel.rows[e.row].dataItem,
        cell: {
          row: e.row,
          col: e.col,
          item: e.panel.rows[e.row].dataItem,
        }
      };

      // For some reason, this now throws an exception. Wrapping in a try/catch since
      // the spreadsheet is still generated regardless.
      try {
        ////////////////////////////////////////////
        // THIS LINE THROWS AN EXCEPTION
        let insTemp = renderToString(templ.template(context));
        e.cell.innerHTML = insTemp;
      } catch (e) {
        console.log(e);
      }
    }
    // if itemFormatter, apply its binding
    if (grid.itemFormatter) {
      grid.itemFormatter(e.panel, e.row, e.col, e.cell);
    }
  });

  clone.invalidate();

  return clone;
}

